//-----------------------------------------------------------------------------
// updater_F02x.c
//-----------------------------------------------------------------------------
// Copyright 2002 Cygnal Integrated Products, Inc.
//
// AUTH: FB
// DATE: 28 JUN 02
//
// This program shows an example Firmware Updater using the 'F02x. It resides
// in FLASH at addresses above 0x1000 and is accessed through a function
// pointer casted as (void code*) 0x1000.   
// 
// Once the firmware update has taken place, the a software reset is issued 
// and the updated firmware takes control of the system. 
// 
// Control Function:
//
// The system is controlled via the hardware UART, operating at a baud rate
// determined by the constant <BAUDRATE>, using Timer1 overflows as the baud
// rate source.
//
// Note: Because this program writes to FLASH, the MONEN pin should be tied
//       high.
//
// Target: C8051F02x
// Tool chain: KEIL C51 6.03 / KEIL EVAL C51
//

#define START 0xe000

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include <c8051f020.h>                 // SFR declarations
#include <stdio.h>                     // printf() and getchar()
#include <stdlib.h>
#include <ctype.h>                     // tolower() and toint()

//-----------------------------------------------------------------------------
// Global CONSTANTS
//-----------------------------------------------------------------------------
#define TRUE         1
#define FALSE        0

#define SYSCLK       22118400          // SYSCLK frequency in Hz
#define BAUDRATE     115200            // Baud rate of UART in bps

/* Clock frequencies */
enum t_clk {
	CLK_2M= 4,
	CLK_4M= 5,
	CLK_8M= 6,
	CLK_16M= 7,
	CLK_11M= 0x18,
	CLK_22M= 8
};

/* Baud-rates */
enum t_br {
	BR_0= 0,
	BR_110= 1,
	BR_150= 2,
	BR_300= 3,
	BR_600= 4,
	BR_1K= 5,
	BR_2K= 6,
	BR_4K= 7,
	BR_9K= 8,
	BR_19K= 9,
	BR_38K= 10,
	BR_56K= 11,
	BR_115K= 12
};

//-----------------------------------------------------------------------------
// Function PROTOTYPES
//-----------------------------------------------------------------------------

void main (void);

// Support Subroutines in Bootloader.c
void print_menu(void);
void erase_flash(void);
void receive_code(void);
unsigned char hex2char();

// Initialization Subroutines
void SYSCLK_Init (void);
void PORT_Init (void);
void UART0_Init (void);

// extension
void setup_keys();
unsigned char pressed(unsigned char mask);
void select_menu();

//-----------------------------------------------------------------------------
// Global VARIABLES
//-----------------------------------------------------------------------------

void (*f)();                           // function pointer declaration

bit code_erased = FALSE;               // flag used to indicate that the FLASH
                                       // erase operation is complete
bit f_valid = FALSE;                   // flag to indicate that the FLASH 
                                       // programming operation is complete
//bit b_menu, b_exit, b_up, b_down;
sbit menu= P3^5;
sbit exit= P3^4;
sbit up=   P3^3;
sbit down= P3^2;
sbit led7= P0^7;
sbit led6= P0^6;
sbit led5= P0^5;
sbit led4= P0^4;

#define m_menu 0x20
#define m_ok   m_menu
#define m_exit 0x10
#define m_esc  m_exit
#define m_up   0x08
#define m_down 0x04

sbit LED0= P0^7;
sbit LED1= P0^6;
sbit LED2= P0^5;
sbit LED3= P0^4;

//-----------------------------------------------------------------------------
// MAIN Routine
//-----------------------------------------------------------------------------

void main (void) 
{
   
	char input;   

    WDTCN = 0xde;                       // disable watchdog timer
    WDTCN = 0xad;

    EA = 0;                             // disable interrupts (this statement
                                       // is needed because the device is not
                                       // neccesarily in a reset state prior 
                                       // to executing this code) 

    PORT_Init ();                       // initialize crossbar and GPIO
    SYSCLK_Init ();                     // initialize oscillator
    UART0_Init ();                      // initialize UART0
   
    print_menu();                       // print the command menu

	while ((P3 & 0xfc) != 0xfc) ;		// Wait to release all buttons
	setup_keys();

    printf("Enter a command > ");
    while (1)
    {
		if (RI0)
	  	{
      		input = getchar();
        	switch ( input )
			{
              	case '1': erase_flash();   
            		printf("\n*** Flash pages erased.\n");
                	receive_code(); 
                                               
         		case '2': printf("\n** RESETTING **\n\n");
            	    RSTSRC = 0x10;      // reset the device                   
         
         		case '?': print_menu(); 
            		break;
                       
         		default:  print_menu(); 
            		printf("\n*** Unknown Command\n");
                	break;
			}
		}
		if (pressed(m_menu))
		{
			select_menu();
		}
	}   

} // end main



static enum t_br used_br;

const long clk_table[]= {
	2000000,
	4500000, //4000000,
	9000000, //8000000,
	18000000,//16000000,
	11059200,
	22118400,
	0
};

const long br_table[]= {
	0, // 0
	110, // 1
	150, // 2
	300, // 3
	600, // 4
	1200, // 5
	2400, // 6
	4800, // 7
	9600, // 8
	19200, // 9
	38400, // 10
	57600, // 11
	115200, // 12
	0
};

long
get_Fosc()
{
	enum t_clk clk= OSCICN&0x7;
	if ((OSCICN & 0x08)!=0)
	{
		/* external */
		if ((OSCXCN & 0x10)!=0)
			return clk_table[4];
		else
			return clk_table[5];
	}
	else
	{
		/* internal */
		switch (clk)
		{
			case CLK_2M:  return clk_table[0]; break;
			case CLK_4M:  return clk_table[1]; break;
			case CLK_8M:  return clk_table[2]; break;
			case CLK_16M: return clk_table[3]; break;
		}
	}
	return 11059200;
}

void
br_init(enum t_br br)
{
	unsigned int r;
	long BR= br_table[br];
	long clk= get_Fosc();

	if (br!=BR_0)
	{
		TR2= 0;
		r= 65536-(clk/(32*BR));
 		RCAP2L= r&0xff;
   		RCAP2H= r/256;
		TL2= RCAP2L;
		TH2= RCAP2H;
		TR2= 1;
	}
	used_br= br;
}

void
oscillator_init(enum t_clk clk)
{
	bit EAorg;
    int i;
	
	EAorg= EA;
	EA= 0;
	i = 0;
	/*
	if ((P3 & 0x30) == 0)
	{
		if ((clk == CLK_22M) || (clk == CLK_11M))
			clk= CLK_16M;
	}
	*/
	if ((clk==CLK_11M) || (clk==CLK_22M))
	{
		/* external */
		OSCXCN= 0x67;
		OSCXCN|= (clk==CLK_11M)?0x10:00;
		OSCICN= (OSCICN&~0x07)|4;
	    for (i = 0; i < 3000; i++);  /* Wait 1ms for initialization */
	    for (i=0; (i<3000)&&((OSCXCN&0x80)!=0x80);i++) ;
		OSCICN= 8|4;
	}
	else
	{
		/* internal */
		OSCICN= (OSCICN&~0x07)|clk;
	    for (i = 0; i < 3000; i++);  /* Wait 1ms for initialization */
	    for (i=0; (i<3000)&&((OSCICN&0x10)!=0x10);i++) ;
		OSCICN&= ~0x08;
		OSCXCN= 7;
	}
	EA= EAorg;
}

void
debounce()
{
	unsigned int i;
	for (i= 0; i<0xffff; i++) ;
}

static unsigned char key_state;

static void
setup_keys()
{
	key_state= P3;
}

static unsigned char
pressed(unsigned char mask)
{
	if ((key_state & mask) != (P3 & mask))
	{
		key_state= (key_state & ~mask) | (P3 & mask);
		debounce();
		return (P3&mask)==0;
	}
	return 0;
}

static void
display3(unsigned char x)
{
	LED0= x&1;
	LED1= x&2;
	LED2= x&4;
}

static void
display4(unsigned char x)
{
	display3(x);
	LED3= x&8;
}

static const enum t_clk clks[]= {
	CLK_2M, /* 0 */
	CLK_4M, /* 1 */
	CLK_8M, /* 2 */
	CLK_16M,/* 3 */
	CLK_11M,/* 4 */
	CLK_22M,/* 5 */
	0
};

static void
do_select_osc()
{
	unsigned char x= 0;
	printf("CLK=%ld\n", get_Fosc());
	while (1)
	{
		display4(x);
		if (pressed(m_up))
		{
			x= (x+1)&15;
			if (clks[x]==0)
				x= 0;
		}
		if (pressed(m_down))
		{
			if (x)
				x= (x-1)&15;
			else
				x= 5;
		}
		if (pressed(m_ok))
		{
			printf("Switching to osc=%d...\n", x);
			oscillator_init(clks[x]);
			P0^= 0xf0;
			br_init(used_br);
			printf("Done, clk=%ld\n", get_Fosc());
			//return;
		}
		if (pressed(m_exit))
		{
			return;
		}
	}
}

unsigned char selected_menu;

void
do_menu()
{
	switch (selected_menu)
	{
		case 0: do_select_osc(); break;
		case 1: break;
		case 2: break;
		case 3: break;
		case 4: break;
		case 5: break;
		case 6: break;
		case 7: break;
	}
}

void
select_menu()
{
	long int l= 0;
	selected_menu= 0;
	printf("\nKeyboard menu selection started\n");
	key_state= P3;
	while (1)
	{
		display3(selected_menu);
		if (pressed(m_menu))
		{
			printf("Starting %c\n", selected_menu+'0');
			do_menu();
			selected_menu= 0;
			P0= 0x0f;
			return;
		}
		if (pressed(m_exit))
		{
			selected_menu= 0;
			P0= 0x0f;
			printf("Exit\n");
			return;
		}
		if (pressed(m_up))
		{
			selected_menu= (selected_menu+1) & 7;
			printf("Selected %c\n", selected_menu+'0');
		}
		if (pressed(m_down))
		{
			selected_menu= (selected_menu-1) & 7;
			printf("Selected %c\n", selected_menu+'0');
		}
		if (l++ > 0x4000)
		{
			l= 0;
			LED3= !LED3;
		}
	}
}


//-----------------------------------------------------------------------------
// Support Subroutines
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// print_menu
//-----------------------------------------------------------------------------
//
// This routine prints the command menu to the UART.
//
void print_menu(void)
{
  
   printf("\n\nC8051F02x Firmware Updater\n");
   printf("---------------------------------\n");
   printf("1. Erase FLASH and Update Firmware\n");
   printf("2. Cancel Firmware Update\n");
   printf("?. Print Command List\n");
   
}

//-----------------------------------------------------------------------------
// hex2char
//-----------------------------------------------------------------------------
//
// This routine converts a two byte ascii representation of a char to an
// 8-bit variable;
//
unsigned char hex2char()
{
   
   unsigned char retval;
   char byteH, byteL;
   
   // get a two-byte ASCII representation of a char from the UART
   byteH = _getkey();
   byteL = _getkey();
   
   // convert to a single 8 bit result
   retval = (char) toint(byteH) * 16;
   retval += (char) toint(byteL);
   return retval;
}

//-----------------------------------------------------------------------------
// erase_flash
//-----------------------------------------------------------------------------
//
// This routine erases the first 8 pages of FLASH (0x0000 to 0x0FFF).
//
void erase_flash(void)
{
   char xdata* data pagePointer = 0;// a pointer to xdata located in data space
                                    // points to the first FLASH page that 
                                    // will be erased
      
   int i;                           // temporary int
   bit EA_state;                    // holds interrupt state

   printf("\n*** Erasing flash from 0x0000 to 0x%04x", START);

   EA_state = EA;                   // save interrupt state

   EA = 0;                          // disable interrupts
   FLSCL |= 0x01;                   // enable FLASH write/erase
   PSCTL  = 0x03;                   // MOVX erases FLASH
   
   // Erase the first 8 FLASH pages
   for (i = 0; i < START/512; i++){
      
      *pagePointer = 0;             // initiate the erase
      
      pagePointer += 512;           // advance to next FLASH page
	  //EA= EA_state;
	if ((i%5) == 0)
	{
	   	printf(".");
		//EA= 0;
		if (P0&0x80)
			P0= 0x0f;
		else
			P0= 0xff;
	}
   }   
	P0= 0x0f;

   PSCTL = 0x00;                    // MOVX writes target XRAM
   FLSCL &= ~0x01;                  // disable FLASH write/erase
   
   EA =  EA_state;                  // restore interrupt state

   f_valid = FALSE;                 // indicate that code is no longer valid
   code_erased = TRUE;              // indicate that FLASH has been erased
}

//-----------------------------------------------------------------------------
// receive_code
//-----------------------------------------------------------------------------
// This routine receives the new firmware through the UART in HEX record 
// format.
//
// Hex Record Format:
//
// +--------+--------+------+-------+--------+------(n bytes)------+----------+
// | RECORD | RECLEN |    OFFSET    | RECORD |                     | CHECKSUM |
// |  MARK  |  (n)   |   (2 BYTES)  |  TYPE  |        DATA         |          |
// |  ':'   |        |              |        |                     |          |
// +--------+--------+------+-------+--------+------(n bytes)------+----------+
//
//
void receive_code(void)
{
   char xdata* data pwrite;               // pointer used for writing FLASH
   char code* data pread;                 // pointer used for reading FLASH
   unsigned char len;                     // holds the HEX record length field
   unsigned char record_type;             // holds the HEX record type field
   unsigned int offset;                   // holds the HEX record offset field
                                          // this is the starting address of
                                          // the code image contained in the 
                                          // record

   char checksum;                         // holds the HEX record checksum field
   char flash_checksum;                   // holds the checksum calculated after
                                          // the FLASH has been programmed
   bit EA_state;                          // temporary holder used to restore 
                                          // interrupts to their previous state

   char c;                                // temporary char
   int i;                                 // temporary int
	int written= 0;

   // make sure FLASH has been erased
   if(!code_erased){
      printf("\n*** At least one FLASH page must be erased prior");
      printf(" to this operation.\n");
      return;
   } else {

      printf("\nReady to receive...\n");
   }

   // wait for the user send HEX file

   do{
      
      while( c = _getkey() != ':' );      // ignore all characters until 
                                          // reaching the record mark field
      
      // get the record length
      len = hex2char();
      
      // get the starting address (offset field in HEX record)
      offset = hex2char();                // get the MSB
      offset <<= 8;                       
      offset |= hex2char();               // get the LSB

      // get the record type
      record_type = hex2char();
      if( record_type != 0 && record_type != 1 ){
         printf("\n*** Cannot decode HEX file.\n"); 
         return;
      }      

      EA_state = EA;                      // save the interrupt enable bit state
      
      EA = 0;                             // disable interrupts (precautionary)   
      FLSCL |= 0x01;                      // enable FLASH write/erase
      PSCTL  = 0x01;                      // MOVX writes FLASH

      pwrite = (char xdata*) offset;      // initialize the write pointer
      
      
      code_erased = FALSE;                // clear the code_erased flag      
           
      // write the record into flash
      for( i = 0; i < len; i++){
         
         // check for valid pointer
         if(pwrite < START){ 
            *pwrite = hex2char();         // write one byte to FLASH
            pwrite++;                     // increment FLASH write pointer
         } else {
            printf("\n\nExceeded Code Space.\n");  // print error message
         }
		 written++;
		 if ((written%256) == 0)
		 {
		 	printf(".");
			P0^= 0x80;
		 }
      }
      P0= 0x0f;
      PSCTL = 0x00;                       // MOVX writes target XRAM
      FLSCL &= ~0x01;                     // disable FLASH write/erase
      EA = EA_state;                      // restore interrupts to previous state
            
      // verify the checksum
      pread =  (char code*) offset;       // initialize the read pointer
      checksum = hex2char();              // get the HEX record checksum field
      flash_checksum = 0;                 // set the flash_checksum to zero

      // add the data field stored in FLASH to the checksum
      for( i = 0; i < len; i++)
      {
         flash_checksum += *pread++;
      } 
      
      // add the remaining fields
      flash_checksum += len;
      flash_checksum += (char) (offset >> 8);
      flash_checksum += (char) (offset & 0x00FF);
      flash_checksum += record_type;
      flash_checksum += checksum;
      
      // verify the checksum (the flash_checksum should equal zero)     
      if(flash_checksum != 0){
         printf("*** Checksum failed, try again");
         return;
      }

   } while(record_type != 1); 
   
   f_valid = TRUE;                        // indicate that download is valid 
   
   printf("\n** Firmware Update Complete. **\n");
}

//-----------------------------------------------------------------------------
// Initialization Subroutines
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// SYSCLK_Init
//-----------------------------------------------------------------------------
//
// This routine initializes the system clock to use an 22.1184MHz crystal
// as its clock source.
//
void SYSCLK_Init (void)
{
   int i;                              // delay counter

   OSCXCN = 0x67;                      // start external oscillator with
                                       // 22.1184MHz crystal

   for (i=0; i < 256; i++) ;           // wait for osc to start

   //while (!(OSCXCN & 0x80)) ;          // Wait for crystal osc. to settle

   OSCICN = 0x88;                      // select external oscillator as SYSCLK
                                       // source and enable missing clock
                                       // detector
}

//-----------------------------------------------------------------------------
// PORT_Init
//-----------------------------------------------------------------------------
//
// Configure the Crossbar and GPIO ports
//
void PORT_Init (void)
{
/*
   XBR0    = 0x04;                     // Enable UART0
   XBR1    = 0x00;
   XBR2    = 0x40;                     // Enable crossbar and weak pull-ups
   P0MDOUT |= 0x01;                    // enable TX0 as a push-pull output
   P1MDOUT |= 0x40;                    // enable P1.6 (LED) as push-pull output
*/
    P0MDOUT   = 0xF0;
    /*P1MDOUT   = 0xFF;*/
    /*P2MDOUT   = 0xFF;*/
    /*P3MDOUT   = 0xFF;*/
    /*P74OUT    = 0xFF;*/
    XBR0      = 0x04;
    XBR2      = 0x40;
    P0= P1= P2= P3= 0xff;
    P4= P5= P6= P7= 0xff;
    P0= 0x0f;
}

//-----------------------------------------------------------------------------
// UART0_Init
//-----------------------------------------------------------------------------
//
// Configure the UART0 using Timer1, for <baudrate> and 8-N-1.
//
void UART0_Init (void)
{
   SCON0   = 0x50;                     // SCON0: mode 1, 8-bit UART, enable RX
   TMOD    = 0x20;                     // TMOD: timer 1, mode 2, 8-bit reload
   TH1    = -(SYSCLK/BAUDRATE/16);     // set Timer1 reload value for baudrate
   TR1    = 1;                         // start Timer1
   CKCON |= 0x10;                      // Timer1 uses SYSCLK as time base
   PCON  |= 0x80;                      // SMOD00 = 1
   TI0    = 1;                         // Indicate TX0 ready
}

